@@ -1,22 +1,22 @@
 /* VM library: PowerPC definitions, to be included from both C and assembly.
 
-   Copyright (C) 2017, 2019, 2020 Luca Saiu
+   Copyright (C) 2017, 2019, 2020, 2021 Luca Saiu
    Written by Luca Saiu
 
-   This file is part of Jitter.
+   This file is part of GNU Jitter.
 
-   Jitter is free software: you can redistribute it and/or modify
+   GNU Jitter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
-   Jitter is distributed in the hope that it will be useful,
+   GNU Jitter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with Jitter.  If not, see <http://www.gnu.org/licenses/>. */
+   along with GNU Jitter.  If not, see <http://www.gnu.org/licenses/>. */
 
 
 #ifndef JITTER_NATIVE_MACHINE_H_ /* One macro suffices for all architectures. */
@@ -183,7 +183,7 @@
  * ************************************************************************** */
 
 /* Ignore this page if patch-ins have been disabled (for debugging) or the
-   dispatching model does not support them. */
+   dispatch does not support them. */
 #if defined(JITTER_MACHINE_SUPPORTS_PATCH_IN) && defined(JITTER_DISPATCH_NO_THREADING)
 
 /* This factors the common logic of low-level conditional fast branches.
@@ -210,7 +210,7 @@
               [jitter_operand0] "r" (opd0),                                 \
               JITTER_INPUT_VM_INSTRUCTION_BEGINNING /* inputs */            \
             : "cr7" /* clobbers */                                          \
-            : jitter_dispatch_label /* goto labels */)
+            : jitter_fake_target /* goto labels */)
 #define _JITTER_LOW_LEVEL_BRANCH_FAST_CONDITIONAL_BINARY_(                  \
            cmp_insn, b_insn, opd0, opd1, tgt)                               \
   asm goto (JITTER_ASM_DEFECT_DESCRIPTOR                                    \
@@ -229,7 +229,7 @@
               [jitter_operand1] "r" (opd1),                                 \
               JITTER_INPUT_VM_INSTRUCTION_BEGINNING /* inputs */            \
             : "cr7" /* clobbers */                                          \
-            : jitter_dispatch_label /* goto labels */)
+            : jitter_fake_target /* goto labels */)
 
 /* Low-level fast branches on sign. */
 #define _JITTER_LOW_LEVEL_BRANCH_FAST_IF_ZERO_(opd0, tgt)  \
@@ -327,7 +327,7 @@
               [jitter_xer_no_overflow] "r" (jitter_xer_no_overflow),            \
               JITTER_INPUT_VM_INSTRUCTION_BEGINNING /* inputs */                \
             : "xer", "cr0" /* clobbers */                                       \
-            : jitter_dispatch_label /* goto labels */);                         \
+            : jitter_fake_target /* goto labels */);                         \
   /* Make sure to get the current value in the register as the result, and not  \
      a previous copy.  In order to force this pretend to update the register    \
      here, only if the branch was not taken.  Inline asm will use the same      \
@@ -408,7 +408,7 @@
               [jitter_operand1] opd1_constraint (opd1),                         \
               JITTER_INPUT_VM_INSTRUCTION_BEGINNING /* inputs */                \
             : JITTER_STRINGIFY (JITTER_SCRATCH_REGISTER), "cr0" /* clobbers */  \
-            : jitter_dispatch_label /* goto labels */)
+            : jitter_fake_target /* goto labels */)
 
 /* This uses the macro above and itself factors two use cases: branch on and,
    branch on not and. */
@@ -499,7 +499,7 @@
 
 /* Return using native machine language facilities, when the return address in
    in the given rvalue. */
-#define JITTER_RETURN(link_rvalue)                                             \
+#define _JITTER_RETURN(link_rvalue)                                            \
   do                                                                           \
     {                                                                          \
       const void * jitter_the_return_address = (const void*) (link_rvalue);    \
@@ -509,7 +509,7 @@
                 : /* outputs. */                                               \
                 : [return_addr] "r" (jitter_the_return_address) /* inputs. */  \
                 : "lr" /* clobbers. */                                         \
-                : jitter_dispatch_label /* gotolabels. */);                    \
+                : jitter_fake_target /* gotolabels. */);                    \
       /* The rest of the VM instruction is unreachable. */                     \
       __builtin_unreachable ();                                                \
     }                                                                          \
@@ -517,7 +517,7 @@
 
 /* Perform a branch-and-link to the pointed callee instruction using specific
    machine language features rather than generic indirect branches. */
-#define JITTER_BRANCH_AND_LINK_INTERNAL(callee_rvalue)                        \
+#define _JITTER_BRANCH_AND_LINK_NATIVE(callee_rvalue)                         \
   do                                                                          \
     {                                                                         \
       const void * const jitter_destination =                                 \
@@ -528,16 +528,17 @@
                 : /* outputs. */                                              \
                 : [destination] "r" (jitter_destination) /* inputs. */        \
                 : "ctr", "lr" /* clobbers. */                                 \
-                : jitter_dispatch_label /* gotolabels. */);                   \
-      /* Skip the rest of the specialised instruction, for compatibility */   \
-      /* with more limited dispatches. */                                     \
-      JITTER_JUMP_TO_SPECIALIZED_INSTRUCTION_END;                             \
+                : jitter_fake_target /* gotolabels. */);                   \
+      /* It would be incorrect to have __builtin_unreachable or               \
+         JITTER_JUMP_TO_SPECIALIZED_INSTRUCTION_END here: see the comment in  \
+         the x86_64 version. */                                               \
     }                                                                         \
   while (false)
 
 /* Branch to the pointed caleee destination, but set the link register to a
    return address different from the natural one. */
-#define JITTER_BRANCH_AND_LINK_WITH(_jitter_callee_rvalue, _jitter_new_link)  \
+#define _JITTER_BRANCH_AND_LINK_WITH_NATIVE(_jitter_callee_rvalue,            \
+                                            _jitter_new_link)                 \
   do                                                                          \
     {                                                                         \
       const void * const jitter_callee_rvalue =                               \
@@ -547,7 +548,10 @@
       asm goto (JITTER_ASM_DEFECT_DESCRIPTOR                                  \
                 JITTER_ASM_COMMENT_UNIQUE("Branch-and-link-with, pretending"  \
                                           "to go to "                         \
-                                          "%l[jitter_dispatch_label]")        \
+                                          "%l[jitter_fake_target]")        \
+                /* This PowerPC version is particularly pretty and natural,   \
+                   and does not negatively affect branch target prediction    \
+                   for once. */                                               \
                 "mtctr %[jitter_callee_rvalue]\n\t"                           \
                 "mtlr %[jitter_new_link]\n\t"                                 \
                 "bctr\n\t"                                                    \
@@ -555,8 +559,8 @@
                 : [jitter_callee_rvalue] "r" (jitter_callee_rvalue),          \
                   [jitter_new_link] "r" (jitter_new_link) /* inputs. */       \
                 : "ctr", "lr" /* clobbers. */                                 \
-                : jitter_dispatch_label /* gotolabels. */);                   \
-      /* The rest of the VM instruction is unreachable: this is an            \
+                : jitter_fake_target /* gotolabels. */);                   \
+      /* The rest of the VM instruction is unreachable: this tail call is an  \
          unconditional jump. */                                               \
       __builtin_unreachable ();                                               \
     }                                                                         \
@@ -565,7 +569,7 @@
 /* Perform an ordinary bl, writing the return address into the link register.
    Of course we need a patch-in, since the destination address is encoded in
    the jumping instruction. */
-#define _JITTER_BRANCH_FAST_AND_LINK_INTERNAL(target_index)                    \
+#define _JITTER_BRANCH_FAST_AND_LINK_NATIVE(target_index)                      \
   do                                                                           \
     {                                                                          \
       asm goto (JITTER_ASM_DEFECT_DESCRIPTOR                                   \
@@ -578,11 +582,10 @@
                 : JITTER_PATCH_IN_INPUTS_FOR_EVERY_CASE,                       \
                   JITTER_INPUT_VM_INSTRUCTION_BEGINNING /* inputs */           \
                 : /* clobbers. */                                              \
-                : jitter_dispatch_label /* gotolabels. */);                    \
-      /* The rest of this specialised instruction is unreachable.  This        \
-         implementation is not based on hardware call and return, so there     \
-         is no need to generate a hardware jump either. */                     \
-      __builtin_unreachable ();                                                \
+                : jitter_fake_target /* gotolabels. */);                    \
+      /* It would be incorrect to have __builtin_unreachable or                \
+         JITTER_JUMP_TO_SPECIALIZED_INSTRUCTION_END here: see the comment in   \
+         the x86_64 version. */                                                \
     }                                                                          \
   while (false)
 
