@@ -588,6 +588,7 @@ my_decl_map_fn (pkl_ast_node decl, void *data)
   char *source =  PKL_AST_DECL_SOURCE (decl);
   int order = PKL_AST_DECL_ORDER (decl);
   char *type = NULL;
+  pk_val value = PK_NULL;
   int kind;
 
   /* Skip mappers, i.e. function declarations whose initials are
@@ -607,6 +608,9 @@ my_decl_map_fn (pkl_ast_node decl, void *data)
       return;
     }
 
+  if (PKL_AST_DECL_KIND (decl) != PKL_AST_DECL_KIND_TYPE)
+    value = pvm_env_lookup (payload->runtime_env, 0, order);
+
   type
     = (PKL_AST_TYPE (initial)
        ? pkl_type_str (PKL_AST_TYPE (initial), 1)
@@ -617,7 +621,7 @@ my_decl_map_fn (pkl_ast_node decl, void *data)
                type,
                loc.first_line, loc.last_line,
                loc.first_column, loc.last_column,
-               pvm_env_lookup (payload->runtime_env, 0, order),
+               value,
                payload->data);
   free (type);
 }
@@ -733,6 +737,7 @@ pk_call (pk_compiler pkc, pk_val cls, pk_val *ret, int narg, ...)
   pvm_program program;
   va_list ap;
   enum pvm_exit_code rret;
+  pvm_val exit_exception;
 
   /* Compile a program that calls the function.  */
   va_start (ap, narg);
@@ -743,10 +748,11 @@ pk_call (pk_compiler pkc, pk_val cls, pk_val *ret, int narg, ...)
 
   /* Run the program in the poke VM.  */
   pvm_program_make_executable (program);
-  rret = pvm_run (pkc->vm, program, ret, NULL);
+  rret = pvm_run (pkc->vm, program, ret, &exit_exception);
 
   pvm_destroy_program (program);
-  PK_RETURN (rret == PVM_EXIT_OK ? PK_OK : PK_ERROR);
+  PK_RETURN (rret == PVM_EXIT_OK && exit_exception == PVM_NULL
+             ? PK_OK : PK_ERROR);
 }
 
 int
