@@ -1,6 +1,6 @@
 /* pkl-trans.c - Transformation phases for the poke compiler.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -224,10 +224,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_offset)
       pkl_ast_node magnitude
         = pkl_ast_make_integer (PKL_PASS_AST, 1);
 
-      PKL_AST_LOC (magnitude_type) = PKL_AST_LOC (offset);
-      PKL_AST_LOC (magnitude) = PKL_AST_LOC (offset);
       PKL_AST_TYPE (magnitude) = ASTREF (magnitude_type);
-
       PKL_AST_OFFSET_MAGNITUDE (offset) = ASTREF (magnitude);
     }
 }
@@ -489,14 +486,16 @@ PKL_PHASE_END_HANDLER
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_op_attr)
 {
   pkl_ast_node exp = PKL_PASS_NODE;
-
-  pkl_ast_node identifier = PKL_AST_EXP_OPERAND (exp, 1);
-  const char *identifier_name = PKL_AST_IDENTIFIER_POINTER (identifier);
+  pkl_ast_node identifier;
+  const char *identifier_name;
   enum pkl_ast_attr attr = PKL_AST_ATTR_NONE;
 
   if (PKL_AST_EXP_ATTR (exp) != PKL_AST_ATTR_NONE)
     PKL_PASS_DONE;
 
+  identifier = PKL_AST_EXP_OPERAND (exp, 1);
+  identifier_name = PKL_AST_IDENTIFIER_POINTER (identifier);
+
   for (attr = 0; pkl_attr_name (attr); ++attr)
     {
       if (STREQ (pkl_attr_name (attr), identifier_name))
@@ -609,9 +608,6 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_trimmer)
 
       from = pkl_ast_make_integer (PKL_PASS_AST, 0);
       PKL_AST_TYPE (from) = ASTREF (idx_type);
-      PKL_AST_LOC (idx_type) = PKL_AST_LOC (trimmer);
-      PKL_AST_LOC (from) = PKL_AST_LOC (trimmer);
-
       PKL_AST_TRIMMER_FROM (trimmer) = ASTREF (from);
     }
 
@@ -643,13 +639,13 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_trimmer)
 }
 PKL_PHASE_END_HANDLER
 
-/* Decode format strings in `printf' instructions.  */
+/* Decode format strings in `format' and `printf'.  */
 
-PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_format)
 {
-  pkl_ast_node print_stmt = PKL_PASS_NODE;
-  pkl_ast_node args = PKL_AST_PRINT_STMT_ARGS (print_stmt);
-  pkl_ast_node print_fmt = PKL_AST_PRINT_STMT_FMT (print_stmt);
+  pkl_ast_node format = PKL_PASS_NODE;
+  pkl_ast_node args = PKL_AST_FORMAT_ARGS (format);
+  pkl_ast_node format_fmt = PKL_AST_FORMAT_FMT (format);
   char *fmt, *p;
   pkl_ast_node t, arg;
   int ntag, nargs = 0;
@@ -659,28 +655,29 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
 #define MAX_CLASS_TAGS 32
   int nclasses = 0;
   char *classes[MAX_CLASS_TAGS];
+  int add_new_percent_arg_p = 0;
+  int add_new_style_arg_p = 0;
+  char *new_style_class = NULL;
 
   /* Calculate the number of arguments.  */
   for (t = args; t; t = PKL_AST_CHAIN (t))
     nargs++;
-  PKL_AST_PRINT_STMT_NARGS (print_stmt) = nargs;
+  PKL_AST_FORMAT_NARGS (format) = nargs;
 
-  /* If this is a `print', or if the format string has been already
-     processed, then we are done.  */
-  if (!print_fmt
-      || PKL_AST_PRINT_STMT_FMT_PROCESSED_P (print_stmt))
+  /* If the format string has been already processed, then we are done.  */
+  if (PKL_AST_FORMAT_FMT_PROCESSED_P (format))
     PKL_PASS_DONE;
 
-  fmt = PKL_AST_STRING_POINTER (print_fmt);
+  fmt = PKL_AST_STRING_POINTER (format_fmt);
   p = fmt;
 
   /* Process the prefix string, if any.  */
   if (*p != '%')
     {
       p = strchrnul (fmt, '%');
-      PKL_AST_PRINT_STMT_PREFIX (print_stmt) = strndup (fmt, p - fmt);
-      if (!PKL_AST_PRINT_STMT_PREFIX (print_stmt))
-        PKL_ICE (PKL_AST_LOC (print_stmt), _("out of memory"));
+      PKL_AST_FORMAT_PREFIX (format) = strndup (fmt, p - fmt);
+      if (!PKL_AST_FORMAT_PREFIX (format))
+        PKL_ICE (PKL_AST_LOC (format), _("out of memory"));
     }
 
   /* Process the format string.  */
@@ -694,10 +691,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
       int prefix = -1;
 
       assert (*p == '%');
-      if (ntag >= nargs && p[1] != '>' && p[1] != '<')
+      if (ntag >= nargs && p[1] != '%' && p[1] != '>' && p[1] != '<')
         {
-          PKL_ERROR (PKL_AST_LOC (print_stmt),
-                     "not enough arguments in printf");
+          PKL_ERROR (PKL_AST_LOC (format), "not enough format arguments");
           PKL_TRANS_PAYLOAD->errors++;
           PKL_PASS_ERROR;
         }
@@ -731,46 +727,47 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
       /* Now process the rest of the tag.  */
       switch (p[1])
         {
+        case '%':
+          p += 2;
+          add_new_percent_arg_p = 1;
+          break;
         case 'v':
           p += 2;
-          PKL_AST_PRINT_STMT_ARG_BASE (arg) = 0; /* Arbitrary.  */
+          PKL_AST_FORMAT_ARG_BASE (arg) = 0; /* Arbitrary.  */
           atype = pkl_ast_make_any_type (PKL_PASS_AST);
-          PKL_AST_PRINT_STMT_ARG_VALUE_P (arg) = 1;
-          PKL_AST_PRINT_STMT_ARG_PRINT_DEPTH (arg)
+          PKL_AST_FORMAT_ARG_VALUE_P (arg) = 1;
+          PKL_AST_FORMAT_ARG_FORMAT_DEPTH (arg)
             = (prefix == -1 ? 0 : prefix);
           switch (flag)
             {
             case 'T':
-              PKL_AST_PRINT_STMT_ARG_PRINT_MODE (arg)
-                = PKL_AST_PRINT_MODE_TREE;
+              PKL_AST_FORMAT_ARG_FORMAT_MODE (arg)
+                = PKL_AST_FORMAT_MODE_TREE;
               break;
             case 'F':
               /* Fallthrough.  */
             case 0:
-              PKL_AST_PRINT_STMT_ARG_PRINT_MODE (arg)
-                = PKL_AST_PRINT_MODE_FLAT;
+              PKL_AST_FORMAT_ARG_FORMAT_MODE (arg)
+                = PKL_AST_FORMAT_MODE_FLAT;
               break;
             default:
               msg = _("invalid flag");
               goto invalid_tag;
             }
-          PKL_AST_LOC (atype) = PKL_AST_LOC (print_fmt);
           types = pkl_ast_chainon (types, atype);
           ntag++;
           break;
         case 's':
           p += 2;
-          PKL_AST_PRINT_STMT_ARG_BASE (arg) = 10; /* Arbitrary.  */
+          PKL_AST_FORMAT_ARG_BASE (arg) = 10; /* Arbitrary.  */
           atype = pkl_ast_make_string_type (PKL_PASS_AST);
-          PKL_AST_LOC (atype) = PKL_AST_LOC (print_fmt);
           types = pkl_ast_chainon (types, atype);
           ntag++;
           break;
         case 'c':
           p += 2;
-          PKL_AST_PRINT_STMT_ARG_BASE (arg) = 256;  /* Arbitrary */
+          PKL_AST_FORMAT_ARG_BASE (arg) = 256;  /* Arbitrary */
           atype = pkl_ast_make_integral_type (PKL_PASS_AST, 8, 0);
-          PKL_AST_LOC (atype) = PKL_AST_LOC (print_fmt);
           types = pkl_ast_chainon (types, atype);
           ntag++;
           break;
@@ -802,12 +799,12 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
 
                 switch (p[base_idx])
                   {
-                  case 'b': PKL_AST_PRINT_STMT_ARG_BASE (arg) = 2; break;
-                  case 'o': PKL_AST_PRINT_STMT_ARG_BASE (arg) = 8; break;
-                  case 'd': PKL_AST_PRINT_STMT_ARG_BASE (arg) = 10; break;
-                  case 'x': PKL_AST_PRINT_STMT_ARG_BASE (arg) = 16; break;
+                  case 'b': PKL_AST_FORMAT_ARG_BASE (arg) = 2; break;
+                  case 'o': PKL_AST_FORMAT_ARG_BASE (arg) = 8; break;
+                  case 'd': PKL_AST_FORMAT_ARG_BASE (arg) = 10; break;
+                  case 'x': PKL_AST_FORMAT_ARG_BASE (arg) = 16; break;
                   case 'c':
-                    PKL_AST_PRINT_STMT_ARG_BASE (arg) = 256;
+                    PKL_AST_FORMAT_ARG_BASE (arg) = 256;
                     if (bits != 8)
                       {
                         msg = _("char format only makes sense with 8 bits");
@@ -819,9 +816,13 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
                     goto invalid_tag;
                   }
 
+                if (bits == 0)
+                  {
+                    msg = _("invalid bit-width");
+                    goto invalid_tag;
+                  }
                 atype = pkl_ast_make_integral_type (PKL_PASS_AST,
                                                     bits, p[1] == 'i');
-                PKL_AST_LOC (atype) = PKL_AST_LOC (print_fmt);
                 types = pkl_ast_chainon (types, atype);
 
                 if (base_idx == 4)
@@ -848,7 +849,6 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
             int end_sc = 0;
             char *class = xmalloc (strlen (fmt) + 1);
             size_t j;
-            pkl_ast_node new_arg;
 
             end_sc = (p[1] == '>');
             p += 2;
@@ -895,53 +895,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
                 class = classes[--nclasses];
               }
 
-            /* Create the new arg and add it to the list of
-               arguments.  */
-            new_arg = pkl_ast_make_print_stmt_arg (PKL_PASS_AST,
-                                                   NULL);
-            PKL_AST_LOC (new_arg) = PKL_AST_LOC (print_fmt);
-
-            if (end_sc)
-              PKL_AST_PRINT_STMT_ARG_END_SC (new_arg) = xstrdup (class);
-            else
-              PKL_AST_PRINT_STMT_ARG_BEGIN_SC (new_arg) = class;
-
-            if (arg)
-              {
-                if (arg == PKL_AST_PRINT_STMT_ARGS (print_stmt))
-                  {
-                    /* Prepend.  */
-                    PKL_AST_CHAIN (new_arg) = arg;
-                    PKL_AST_PRINT_STMT_ARGS (print_stmt)
-                      = ASTREF (new_arg);
-                  }
-                else
-                  {
-                    /* Add after.  */
-                    PKL_AST_CHAIN (new_arg) = PKL_AST_CHAIN (prev_arg);
-                    PKL_AST_CHAIN (prev_arg) = ASTREF (new_arg);
-                  }
-              }
-            else
-              {
-                /* Append.  */
-                if (!PKL_AST_PRINT_STMT_ARGS (print_stmt))
-                  PKL_AST_PRINT_STMT_ARGS (print_stmt)
-                    = ASTREF (new_arg);
-                else
-                  PKL_AST_PRINT_STMT_ARGS (print_stmt)
-                    = pkl_ast_chainon (PKL_AST_PRINT_STMT_ARGS (print_stmt),
-                                       new_arg);
-              }
-
-            arg = new_arg;
-
-            /* The type corresponding to a styling class format
-               directive is `void'.  */
-            atype = pkl_ast_make_void_type (PKL_PASS_AST);
-            PKL_AST_LOC (atype) = PKL_AST_LOC (print_fmt);
-            types = pkl_ast_chainon (types, atype);
-
+            assert (new_style_class == NULL);
+            new_style_class = end_sc ? xstrdup (class) : class;
+            add_new_style_arg_p = 1;
             break;
           }
         default:
@@ -949,13 +905,80 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
           goto invalid_tag;
         }
 
+      assert (!(add_new_percent_arg_p && add_new_style_arg_p));
+      if (add_new_percent_arg_p || add_new_style_arg_p)
+        {
+          pkl_ast_node new_arg;
+
+          /* Create the new arg and add it to the list of arguments.  */
+          new_arg = pkl_ast_make_format_arg (PKL_PASS_AST, NULL);
+
+          if (add_new_percent_arg_p)
+            PKL_AST_FORMAT_ARG_SUFFIX (new_arg) = xstrdup ("%");
+          else if (add_new_style_arg_p)
+            {
+              int end_sc = p[-1] == '>';
+
+              assert (new_style_class != NULL);
+              if (end_sc)
+                PKL_AST_FORMAT_ARG_END_SC (new_arg) = new_style_class;
+              else
+                PKL_AST_FORMAT_ARG_BEGIN_SC (new_arg) = new_style_class;
+              new_style_class = NULL;
+            }
+
+          if (arg)
+            {
+              if (arg == PKL_AST_FORMAT_ARGS (format))
+                {
+                  /* Prepend.  */
+                  PKL_AST_CHAIN (new_arg) = arg;
+                  PKL_AST_FORMAT_ARGS (format) = ASTREF (new_arg);
+                }
+              else
+                {
+                  /* Add after.  */
+                  PKL_AST_CHAIN (new_arg) = PKL_AST_CHAIN (prev_arg);
+                  PKL_AST_CHAIN (prev_arg) = ASTREF (new_arg);
+                }
+            }
+          else
+            {
+              /* Append.  */
+              if (!PKL_AST_FORMAT_ARGS (format))
+                PKL_AST_FORMAT_ARGS (format) = ASTREF (new_arg);
+              else
+                PKL_AST_FORMAT_ARGS (format)
+                  = pkl_ast_chainon (PKL_AST_FORMAT_ARGS (format), new_arg);
+            }
+
+          arg = new_arg;
+
+          /* The type corresponding to new arg is `void'.  */
+          atype = pkl_ast_make_void_type (PKL_PASS_AST);
+          types = pkl_ast_chainon (types, atype);
+
+          add_new_percent_arg_p = 0;
+          add_new_style_arg_p = 0;
+        }
+
       /* Add the optional suffix to the argument.  */
       if (*p != '\0' && *p != '%')
         {
           char *end = strchrnul (p, '%');
-          PKL_AST_PRINT_STMT_ARG_SUFFIX (arg) = strndup (p, end - p);
-          if (!PKL_AST_PRINT_STMT_ARG_SUFFIX (arg))
-            PKL_ICE (PKL_AST_LOC (print_stmt), _("out of memory"));
+          if (PKL_AST_FORMAT_ARG_SUFFIX (arg))
+            {
+              char* s = NULL;
+
+              if (asprintf(&s, "%s%.*s", PKL_AST_FORMAT_ARG_SUFFIX (arg),
+                           (int)(end - p), p) == -1)
+                PKL_ICE (PKL_AST_LOC (format), _("out of memory"));
+              PKL_AST_FORMAT_ARG_SUFFIX (arg) = s;
+            }
+          else
+            PKL_AST_FORMAT_ARG_SUFFIX (arg) = strndup (p, end - p);
+          if (!PKL_AST_FORMAT_ARG_SUFFIX (arg))
+            PKL_ICE (PKL_AST_LOC (format), _("out of memory"));
           p = end;
         }
     }
@@ -969,19 +992,18 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_print_stmt)
 
   if (nargs > ntag)
     {
-      PKL_ERROR (PKL_AST_LOC (print_stmt),
-                 "too many arguments in printf");
+      PKL_ERROR (PKL_AST_LOC (format), "too many format arguments");
       PKL_TRANS_PAYLOAD->errors++;
       PKL_PASS_ERROR;
     }
 
-  PKL_AST_PRINT_STMT_TYPES (print_stmt) = ASTREF (types);
+  PKL_AST_FORMAT_TYPES (format) = ASTREF (types);
 
-  PKL_AST_PRINT_STMT_FMT_PROCESSED_P (print_stmt) = 1;
+  PKL_AST_FORMAT_FMT_PROCESSED_P (format) = 1;
   PKL_PASS_DONE;
 
  invalid_tag:
-  PKL_ERROR (PKL_AST_LOC (print_fmt),
+  PKL_ERROR (PKL_AST_LOC (format_fmt),
              "invalid %%- tag in format string: %s", msg);
   PKL_TRANS_PAYLOAD->errors++;
   PKL_PASS_ERROR;
@@ -1015,16 +1037,11 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_array)
         {
           pkl_ast_node initializer_index_type
             = pkl_ast_make_integral_type (PKL_PASS_AST, 64, 0);
-          PKL_AST_LOC (initializer_index_type)
-            = PKL_AST_LOC (tmp);
-
 
           initializer_index_node
             = pkl_ast_make_integer (PKL_PASS_AST, index);
           PKL_AST_TYPE (initializer_index_node)
             = ASTREF (initializer_index_type);
-          PKL_AST_LOC (initializer_index_node)
-            = PKL_AST_LOC (tmp);
 
           PKL_AST_ARRAY_INITIALIZER_INDEX (tmp)
             = ASTREF (initializer_index_node);
@@ -1037,8 +1054,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_array)
           if (PKL_AST_CODE (initializer_index_node)
               != PKL_AST_INTEGER)
             {
-              PKL_ICE (PKL_AST_NOLOC,
-                       "array initialize index should be an integer node");
+              PKL_ERROR (PKL_AST_LOC (initializer_index_node),
+                         "indexes in array initializers shall be constant");
+              PKL_TRANS_PAYLOAD->errors++;
               PKL_PASS_ERROR;
             }
 
@@ -1080,8 +1098,8 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_comp_stmt)
 }
 PKL_PHASE_END_HANDLER
 
-/* FOR-IN statements introduce a lexical level if they use an
-   iterator.  Update the function back.  */
+/* FOR-IN statements introduce a lexical level if they use an iterator
+   or have a head of declarations.  Update the function back accordingly.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_loop_stmt_iterator)
 {
@@ -1089,12 +1107,23 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_loop_stmt_iterator)
 }
 PKL_PHASE_END_HANDLER
 
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_loop_stmt)
+{
+  pkl_ast_node stmt = PKL_PASS_NODE;
+
+  if (PKL_AST_LOOP_STMT_HEAD (stmt))
+    {
+      assert (!PKL_AST_LOOP_STMT_ITERATOR (stmt));
+      PKL_TRANS_INCR_FUNCTION_BACK;
+    }
+}
+PKL_PHASE_END_HANDLER
+
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_loop_stmt)
 {
   pkl_ast_node stmt = PKL_PASS_NODE;
 
-
-  if (PKL_AST_LOOP_STMT_ITERATOR (stmt))
+  if (PKL_AST_LOOP_STMT_ITERATOR (stmt) || PKL_AST_LOOP_STMT_HEAD (stmt))
     PKL_TRANS_DECR_FUNCTION_BACK;
 }
 PKL_PHASE_END_HANDLER
@@ -1163,13 +1192,14 @@ struct pkl_phase pkl_phase_trans1 =
    PKL_PHASE_PS_HANDLER (PKL_AST_VAR, pkl_trans1_ps_var),
    PKL_PHASE_PS_HANDLER (PKL_AST_FUNC, pkl_trans1_ps_func),
    PKL_PHASE_PS_HANDLER (PKL_AST_TRIMMER, pkl_trans1_ps_trimmer),
-   PKL_PHASE_PS_HANDLER (PKL_AST_PRINT_STMT, pkl_trans1_ps_print_stmt),
+   PKL_PHASE_PS_HANDLER (PKL_AST_FORMAT, pkl_trans1_ps_format),
    PKL_PHASE_PR_HANDLER (PKL_AST_DECL, pkl_trans1_pr_decl),
    PKL_PHASE_PS_HANDLER (PKL_AST_DECL, pkl_trans1_ps_decl),
    PKL_PHASE_PS_HANDLER (PKL_AST_ARRAY, pkl_trans1_ps_array),
    PKL_PHASE_PR_HANDLER (PKL_AST_COMP_STMT, pkl_trans1_pr_comp_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_COMP_STMT, pkl_trans1_ps_comp_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_LOOP_STMT_ITERATOR, pkl_trans1_ps_loop_stmt_iterator),
+   PKL_PHASE_PR_HANDLER (PKL_AST_LOOP_STMT, pkl_trans1_pr_loop_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_LOOP_STMT, pkl_trans1_ps_loop_stmt),
    PKL_PHASE_PR_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_trans1_pr_struct_type_field),
    PKL_PHASE_PS_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_trans1_ps_struct_type_field),
@@ -1355,8 +1385,6 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_type_offset)
   /* Calculate the size of the complete type in bytes and put it in
      an integer node.  */
   unit = pkl_ast_sizeof_type (PKL_PASS_AST, unit_type);
-  PKL_AST_LOC (unit) = PKL_AST_LOC (unit_type);
-  PKL_AST_LOC (PKL_AST_TYPE (unit)) = PKL_AST_LOC (unit_type);
 
   /* Replace the unit type with this expression.  */
   PKL_AST_TYPE_O_UNIT (type) = ASTREF (unit);
@@ -1386,7 +1414,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_incrdecr)
       /* Get the step.  The type of the expression is safe as per
          typify.  */
       step = pkl_ast_type_incr_step (PKL_PASS_AST, incrdecr_exp_type);
-      assert (step); /* XXX turn to ICE.  */
+      if (!step)
+        PKL_ICE (PKL_AST_NOLOC,
+                 "pkl_ast_type_incr_step failed in pkl_trans2_ps_incrdecr");
 
       /* Build a statement EXP = EXP +/- STEP  */
       exp_plus_one = pkl_ast_make_binary_exp (PKL_PASS_AST, op,
@@ -1401,6 +1431,98 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_incrdecr)
 }
 PKL_PHASE_END_HANDLER
 
+/* Calculate the size of struct type fields that are complete.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_struct_type_field)
+{
+  pkl_ast_node field = PKL_PASS_NODE;
+  pkl_ast_node field_type = PKL_AST_STRUCT_TYPE_FIELD_TYPE (field);
+
+  if (pkl_ast_type_is_complete (field_type) == PKL_AST_TYPE_COMPLETE_YES)
+    {
+      PKL_AST_STRUCT_TYPE_FIELD_SIZE (field)
+        = ASTREF (pkl_ast_sizeof_type (PKL_PASS_AST, field_type));
+      PKL_PASS_RESTART = 1;
+    }
+}
+PKL_PHASE_END_HANDLER
+
+/* Compute the attributes of values with complete types.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_op_attr)
+{
+  pkl_ast_node exp = PKL_PASS_NODE;
+  pkl_ast_node exp_type = PKL_AST_TYPE (exp);
+  pkl_ast_node operand = PKL_AST_EXP_OPERAND (exp, 0);
+  pkl_ast_node operand_type = PKL_AST_TYPE (operand);
+
+  switch (PKL_AST_EXP_ATTR (exp))
+    {
+    case PKL_AST_ATTR_LENGTH:
+      {
+        size_t length = 0;
+        pkl_ast_node len;
+
+        if (PKL_AST_LITERAL_P (operand))
+          switch (PKL_AST_CODE (operand))
+            {
+            case PKL_AST_STRING:
+              length = PKL_AST_STRING_LENGTH (operand);
+              break;
+            case PKL_AST_ARRAY:
+              length = PKL_AST_ARRAY_NELEM (operand);
+              break;
+            default:
+              PKL_PASS_DONE;
+            }
+        else if (PKL_AST_TYPE_CODE (operand_type) == PKL_TYPE_STRUCT
+                 && (PKL_AST_TYPE_COMPLETE (operand_type)
+                     == PKL_AST_TYPE_COMPLETE_YES))
+          length = PKL_AST_TYPE_S_NELEM (operand_type);
+        else
+          PKL_PASS_DONE;
+
+        /* The type of 'length is uint<64>  */
+
+        len = pkl_ast_make_integer (PKL_PASS_AST, length);
+        PKL_AST_TYPE (len) = ASTREF (exp_type);
+        PKL_PASS_NODE = ASTREF (len);
+        pkl_ast_node_free (exp);
+        PKL_PASS_RESTART = 1;
+        PKL_PASS_DONE;
+      }
+    case PKL_AST_ATTR_SIZE:
+      {
+        pkl_ast_node size, off;
+
+        if (PKL_AST_LITERAL_P (operand)
+            && PKL_AST_CODE (operand) == PKL_AST_STRING)
+          {
+            size = pkl_ast_make_integer (PKL_PASS_AST,
+                                         (PKL_AST_STRING_LENGTH (operand) + 1)
+                                           * 8);
+            PKL_AST_TYPE (size) = PKL_AST_TYPE_O_BASE_TYPE (exp_type);
+          }
+        else if (PKL_AST_TYPE_COMPLETE (operand_type)
+                   == PKL_AST_TYPE_COMPLETE_YES)
+          size = pkl_ast_sizeof_type (PKL_PASS_AST, operand_type);
+        else
+          PKL_PASS_DONE;
+
+        off = pkl_ast_make_offset (PKL_PASS_AST, size,
+                                   PKL_AST_TYPE_O_UNIT (exp_type));
+        PKL_AST_TYPE (off) = ASTREF (exp_type);
+        PKL_PASS_NODE = ASTREF (off);
+        pkl_ast_node_free (exp);
+        PKL_PASS_RESTART = 1;
+        PKL_PASS_DONE;
+      }
+    default:
+      break;
+    }
+}
+PKL_PHASE_END_HANDLER
+
 struct pkl_phase pkl_phase_trans2 =
   {
    PKL_PHASE_PS_HANDLER (PKL_AST_SRC, pkl_trans_ps_src),
@@ -1415,6 +1537,8 @@ struct pkl_phase pkl_phase_trans2 =
    PKL_PHASE_PS_HANDLER (PKL_AST_CAST, pkl_trans2_ps_cast),
    PKL_PHASE_PS_HANDLER (PKL_AST_INCRDECR, pkl_trans2_ps_incrdecr),
    PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_OFFSET, pkl_trans2_ps_type_offset),
+   PKL_PHASE_PS_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_trans2_ps_struct_type_field),
+   PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_ATTR, pkl_trans2_ps_op_attr),
   };
 
 
@@ -1442,24 +1566,18 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans3_ps_op_sizeof)
        an integer node.  */
     pkl_ast_node magnitude
       = pkl_ast_sizeof_type (PKL_PASS_AST, op);
-    PKL_AST_LOC (magnitude) = PKL_AST_LOC (node);
-    PKL_AST_LOC (PKL_AST_TYPE (magnitude)) = PKL_AST_LOC (node);
 
     /* Build an offset with that magnitude, and unit bits.  */
     unit_type = pkl_ast_make_integral_type (PKL_PASS_AST, 64, 0);
-    PKL_AST_LOC (unit_type) = PKL_AST_LOC (node);
 
     unit = pkl_ast_make_integer (PKL_PASS_AST, PKL_AST_OFFSET_UNIT_BITS);
-    PKL_AST_LOC (unit) = PKL_AST_LOC (node);
     PKL_AST_TYPE (unit) = ASTREF (unit_type);
 
     offset = pkl_ast_make_offset (PKL_PASS_AST, magnitude, unit);
 
-    PKL_AST_LOC (offset) = PKL_AST_LOC (node);
     offset_type = pkl_ast_make_offset_type (PKL_PASS_AST,
                                             PKL_AST_TYPE (magnitude),
                                             unit);
-    PKL_AST_LOC (offset_type) = PKL_AST_LOC (node);
     PKL_AST_TYPE (offset) = ASTREF (offset_type);
   }
 
