@@ -1,6 +1,6 @@
 /* pkl.h - Poke compiler.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -74,43 +74,52 @@
    RT_PATH should contain the name of a directory where the compiler can
    find its run-time support files.
 
+   FLAGS is an ORed value of PKL_F_* values, or 0.
+
    If there is an error creating the compiler this function returns
    NULL.  */
 
-pkl_compiler pkl_new (pvm vm, const char *rt_path);
+#define PKL_F_NOSTDTYPES 1 /* Do not define standard types.  */
+
+pkl_compiler pkl_new (pvm vm, const char *rt_path, uint32_t flags);
 
 void pkl_free (pkl_compiler compiler);
 
 /* Compile an execute a Poke program from the given file FNAME.
    Return 1 if the compilation was successful, 0 otherwise.
 
-   If EXIT_STATUS is not NULL, set it to the status in which the
-   executed program terminated.  */
+   If not NULL, *EXIT_EXCEPTION is set to an exception value if the
+   execution of the program gets interrupted by an unhandled exception.
+   Otherwise *EXIT_EXCEPTION is set to PK_NULL.  */
 
 int pkl_execute_file (pkl_compiler compiler, const char *fname,
-                      int *exit_status);
+                      pvm_val *exit_exception);
 
 /* Compile and execute Poke program from a NULL-terminated string
    BUFFER.  Return 0 in case of a compilation error, 1 otherwise.  If
    not NULL, END is set to the first character in BUFFER that is not
-   part of the compiled entity.  */
+   part of the compiled entity.
+
+   If not NULL, *EXIT_EXCEPTION is set to an exception value if the
+   execution of the program gets interrupted by an unhandled exception.
+   Otherwise *EXIT_EXCEPTION is set to PK_NULL.  */
 
 int pkl_execute_buffer (pkl_compiler compiler, const char *buffer,
-                        const char **end);
+                        const char **end, pvm_val *exit_exception);
 
 /* Like pkl_execute_buffer, but compile and execute a single Poke
    expression, that generates a value in VAL. */
 
 int pkl_execute_expression (pkl_compiler compiler,
                             const char *buffer, const char **end,
-                            pvm_val *val);
+                            pvm_val *val, pvm_val *exit_exception);
 
 /* Like pkl_execute_expression but compile and execute a single Poke statement,
    which may generate a value in VAL if it is an "expression
    statement".  Otherwise VAL is set to PVM_NULL.  */
 
 int pkl_execute_statement (pkl_compiler compiler, const char *buffer, const char **end,
-                           pvm_val *val);
+                           pvm_val *val, pvm_val *exit_exception);
 
 /* Compile a single Poke expression and return the resulting PVM
    program.  */
@@ -125,14 +134,16 @@ pvm_program pkl_compile_expression (pkl_compiler compiler,
    RET is set to the value returned by the function, or to PK_NULL if
    it is a void function.
 
-   AP contains a list of PVM values to be passed to the function as
-   arguments, terminated by PVM_NULL.
+   NARG is the number of actuals to pass to the function.
+
+   AP contains a list of NARG PVM values to be passed to the function
+   as actual arguments.
 
    Return the compiled PVM program, or NULL if there is a problem
    performing the operation.  */
 
 pvm_program pkl_compile_call (pkl_compiler compiler, pvm_val cls, pvm_val *ret,
-                              va_list ap);
+                              int narg, va_list ap);
 
 /* Return the VM associated with COMPILER.  */
 
@@ -174,8 +185,43 @@ void pkl_set_quiet_p (pkl_compiler compiler, int quiet_p);
 
 /* Get/install a handler for alien tokens.  */
 
-typedef char *(*pkl_alien_token_handler_fn) (const char *id,
-                                             char **errmsg);
+#define PKL_ALIEN_TOKEN_IDENTIFIER 0
+#define PKL_ALIEN_TOKEN_INTEGER    1
+#define PKL_ALIEN_TOKEN_OFFSET     2
+#define PKL_ALIEN_TOKEN_STRING     3
+
+struct pkl_alien_token
+{
+  int kind;
+  union
+  {
+    char *identifier;
+
+    struct
+    {
+      uint64_t magnitude;
+      int width;
+      int signed_p;
+    } integer;
+
+    struct
+    {
+      uint64_t magnitude;
+      int width;
+      int signed_p;
+      uint64_t unit;
+    } offset;
+
+    struct
+    {
+      const char *str;
+    } string;
+
+  } value;
+};
+
+typedef struct pkl_alien_token *(*pkl_alien_token_handler_fn) (const char *id,
+                                                               char **errmsg);
 
 pkl_alien_token_handler_fn pkl_alien_token_fn (pkl_compiler compiler);
 
@@ -203,8 +249,8 @@ char *pkl_resolve_module (pkl_compiler compiler, const char *module,
                           int filename_p);
 
 /* Load a module using the given compiler.
-   If the module cannot be loaded, return 1.
-   Otherwise, return 0.  */
+   If the module cannot be loaded, return 0.
+   Otherwise, return 1.  */
 
 int pkl_load (pkl_compiler compiler, const char *module);
 
